/*
 *      ISIS release V1.1, Dec. 1988
 *      Export restrictions apply
 */
/*
 *  Task.c -- code related to task management
 */

# include <string.h>

# include "isis.h"


task    isis_scheduler;              /* Task descriptor for system task */      
task    *isis_ctp = &isis_scheduler;      /* Scheduler is running initially */
static  task *old_tp;

/* On some machine setjmp and longjmp work fine for ISIS */
#ifndef THREADS
#    if     (GOULD|hp9000s800|NEXT)
#        define    isis_setjmp      _setjmp
#        define    isis_longjmp     _longjmp
#    endif  (GOULD|hp9000s800|NEXT)
#    ifdef  SUN4
#        include   "jmp_pragmas.h"
#    endif
#else  THREADS
#     define   isis_setjmp(x)    0
#     define   isis_longjmp(x)   panic("isis_longjmp shouldn't have been called")

#endif THREADS


static  message *task_msg_arg;
int     TASKLEN;
address NULLADDRESS;
static  entry_stacksize[128];
#define	esize(n)	(n >= 0 && n <= 127 && entry_stacksize[n] > 0)
         
static adesc   st_ad ={ 0, 0, 4};

# define        st_alloc()      ((task*)mallocate(&st_ad))
# define        st_free(x)      mdeallocate((char*)x, &st_ad)

t_init()
  {
	if(isis_state&ISIS_TINIT)
	    return;
	isis_state |= ISIS_TINIT;
	if(isis_tasks == NULLQP)
	    isis_tasks = qu_null();
#ifdef THREADS
	MUTEX_ALLOC(isis_mutex);
        MUTEX_LOCK(isis_mutex);
	CONDITION_ALLOC(isis_ctp->task_runme);
	THREAD_SELF(isis_ctp->task_threadid);
#endif THREADS
        if(TASKLEN == 0)
            t_set_stacksize(DEF_TASKLEN); 
  }
 
#define task_enqueue(cp, tp, discipline)                        \
  {                                                             \
        if(tp)                                                  \
        {                                                       \
            register qnode *qp;                                 \
            if((qp = *cp) == 0)                                 \
            {                                                   \
                qu_tnull(qp);                                   \
                *cp = qp;                                       \
            }                                                   \
            if(discipline == SU)                                \
                qu_add_tp(qp->qu_next, TA_ISTASK, tp);          \
            else                                                \
                qu_add_tp(qp, TA_ISTASK, tp);                   \
            tp->task_queue = cp;                                \
        }                                                       \
  }

task *
do_task_dequeue(cp)
  register condition *cp;
  {
        register qnode *qp, *np;
        register task *tp;
        qp = *cp;
        np = qu_head(qp);
        tp = np->qu_task;
        tp->task_queue = 0;
        qu_free(np);
        if(qp->qu_next == qp)
        {
            qu_free(qp);
            *cp = 0;
        }
        return(tp);
  }

static	short use_lwpstk;

t_set_stacksize(size) 
  register size;
  { 
#       if(THREADS)
#           if(SUNLWP)
                if(size == -1)
		{
		    ++use_lwpstk;
		    return;
		}
                size += sizeof(task);
#           else
                size = sizeof(task);
#           endif
#       else
            size += sizeof(task);
#       endif
        if(TASKLEN > size)
            return;
        st_ad.a_isize = TASKLEN = size; 
  }

isis_entry_stacksize(entry, size)
  {
#ifndef	THREADS
        entry_stacksize[entry] = size+sizeof(task);
#else
#ifdef  SUNLWP
        if(size != -1)
	    size += sizeof(task);
        entry_stacksize[entry] = size;
#endif
#endif	THREADS
  }

t_scheck()
  {
#ifndef THREADS
        char stackp;
        if(&stackp <= isis_ctp->task_data && isis_ctp != &isis_scheduler)
            spanic();
#endif THREADS
  }

spanic()
  {
        int stackp;
        print("ISIS stack overflow for task ");
        cl_rname(isis_ctp->task_routine);
        print("(%x) qu_freelist %x\n", isis_ctp->task_arg0, qu_freelist);
        panic("ISIS stack overflow");
  }

#ifndef THREADS
#define T_scheck()                                                      \
  {                                                                     \
        char stackp;                                                    \
        if(&stackp <= isis_ctp->task_data && isis_ctp != &isis_scheduler)\
            spanic();                                                   \
  }
#else
#define T_scheck()
#endif THREADS

isis_fork_urgent(routine, arg0, mp)
  ifunc *routine;
  char *arg0;
  message *mp;
  {
        ISIS_ENTER();
        if(task_msg_arg = mp)
            msg_increfcount(mp);
        task_enqueue(&isis_runqueue, isis_ctp, SU);
        task_swtch(NULLTASK, routine, arg0);
        ISIS_EXIT();
  }

static  task *delayed_task, *new_task, *T_fork();

task *
isis_fork(routine, arg0, mp)
  ifunc *routine;
  char *arg0;
  message *mp;
  {
	ISIS_ENTER();
        if(task_msg_arg = mp)
            msg_increfcount(mp);
        delayed_task = isis_ctp;
        task_swtch(NULLTASK, routine, arg0);
        ISIS_RETURN(new_task);
  }

static delayedcall(routine, arg0)
  ifunc *routine;
  char *arg0;
  {
        register task *tp = delayed_task;
        delayed_task = 0;
        new_task = isis_ctp;
        task_enqueue(&isis_runqueue, isis_ctp, SD);
        task_swtch(tp, NULLROUTINE, NULLARG);
        if(isis_state&ISIS_XBYREF)
	    (*routine)(&arg0);
	else
            (*routine)(arg0);
  }

static  char *stackp;
static  ifunc *call_on_sys_stack;
static  char *call_arg;
static  condition call_wait;

t_on_sys_stack(routine, arg)
  ifunc *routine;
  char *arg;
  {
        register res;
	ISIS_ENTER();
        if(isis_ctp == &isis_scheduler)
            ISIS_RETURN( (*routine)(arg) );
        call_on_sys_stack = routine;
        call_arg = arg;
        res = t_wait_l(&call_wait, "isis system: waiting for a call on sys stack to finish");
        ISIS_RETURN(res);
  }

static task *free_tp = 0;

task_swtch(tp, routine, arg0)
  register task *tp;
  register ifunc *routine;
  register char *arg0;
  {
	extern invoke();
        if(call_on_sys_stack)
            panic("Illegal attempt to wait while running on system stack");
        T_scheck();
        if(isis_state&ISIS_XBYREF)
            isis_ctp->task_flag |= TASK_XBYREF;
        else
            isis_ctp->task_flag &= ~TASK_XBYREF;
#ifndef	THREADS
        /* Save entry information */
        if(isis_setjmp(isis_ctp->task_env))
            goto task_resumes;
#endif
        /* Allocate a stack for the new task if needed */
        old_tp = isis_ctp;
        if(tp == 0)
        {
	    register STACKLEN;
            if((isis_state&ISIS_RECEIVING) && esize(isis_enum))
                tp = (task*)malloc(STACKLEN = entry_stacksize[isis_enum]);
            else
	    {
                tp = st_alloc();
                STACKLEN = TASKLEN;
	    }
	    STACKLEN -= sizeof(task) + (1 words); 
            ++isis_created;
            tp->task_queue = 0;
            tp->task_msgid = 0;
            tp->task_cond = 0;
            tp->task_waitingfor = 0;
            tp->task_mwant = 0;
            tp->task_msg = 0;
            tp->task_flag = isis_ctp->task_flag&ISIS_XBYREF;
            tp->task_act = isis_ctp->task_act;
            bclr(tp->task_watching);
            tp->task_msgs = qu_null();
            tp->task_active = qu_add_tp(isis_tasks, TA_ISTASK, tp);
            tp->task_routine = routine;
            tp->task_arg0 = arg0;
            tp->task_addr = my_address;
	    tp->task_entry = isis_enum;
	    tp->task_addr.entry = isis_enum;
	    isis_enum = 0;
            tp->task_msg_arg = task_msg_arg;
            task_msg_arg = 0;
            tp->task_cohorts = (address*)0;
            tp->task_data = ((char*)(&tp->task_data))+sizeof(char*);
	    tp->task_end = (char*)&tp->task_data[STACKLEN];
#ifdef      SUNLWP
            {
		register n = tp->task_entry;
                if(use_lwpstk || (n >= 0 && n <= 127 && entry_stacksize[n] == -1))
                    tp->task_end = tp->task_data = (char*)lwp_newstk();
	    }
#endif      SUNLWP
#ifdef  THREADS
	    CONDITION_ALLOC(tp->task_runme);
	    THREAD_FORK(invoke, tp, tp->task_end);
#else   THREADS
            /* Change stack pointers.  Alignment is important for many machines */ 
#           if(VAX)
            {
                stackp = &tp->task_data[STACKLEN-(2 words)];
                stackp = (char*)((int)stackp & ~0x3);
                asm("movl  _stackp,sp");
            }   
#           endif
#           if(GOULD)
            {
                stackp = &tp->task_data[STACKLEN-(24 words)];
                stackp = (char*)((int)stackp & ~0x7);
                asm("movw  _stackp,b2");
            }   
#           endif
#           if(SUN3)
            {
                stackp = &tp->task_data[STACKLEN-(2 words)];
                stackp = (char*)((int)stackp & ~0x3);
                asm("movl  _stackp,sp");
            }   
#           endif
#           if(NEXT)
            {
                stackp = &tp->task_data[STACKLEN-(16 words)];
                stackp = (char*)((int)stackp & ~0x3);
                asm("movel  _stackp,sp");
            }   
#           endif
#           if(SUN4)
            {
#               include <sun4/asm_linkage.h>
                stackp = &tp->task_data[STACKLEN-(32 words)];
                stackp = (char*)((int)stackp & ~(STACK_ALIGN-1));
                asm("sethi   %hi(_stackp),%o0");
                asm("ld      [%o0+%lo(_stackp)],%sp");
            }   
#           endif
#           if(HPUX)
            {
#ifdef          hp9000s300
                    stackp = &tp->task_data[STACKLEN-(2 words)];
                    stackp = (char*)((int)stackp & ~0x3);
                    asm("mov.l  _stackp,%sp");
#endif          hp9000s300
#ifdef          hp9000s800
                    stackp = &tp->task_data[0];
                    stackp = (char*)((((int)stackp+7) & ~0x7)+ 48);
                    set_sp(stackp);
#endif          hp9000s800

            }
#           endif
            invoke(tp);
            /* In this case the child task is finishing up */
            tp = task_dequeue(&isis_runqueue);
#endif  THREADS
        }
#ifdef  THREADS
        else
	    CONDITION_SIGNAL(tp->task_runme);
#endif  THREADS
        ++isis_switched;
	isis_ctp = tp;
#ifdef	THREADS
        CONDITION_WAIT(old_tp->task_runme);
#else 	THREADS
        isis_longjmp(tp->task_env, 1);
task_resumes:
#endif	THREADS
        if(isis_ctp->task_flag&TASK_XBYREF)
	    isis_state |= ISIS_XBYREF;
        else
	    isis_state &= ~ISIS_XBYREF;
        if(free_tp)
        {
	    if(free_tp->task_entry != -1 && esize(free_tp->task_entry))
	        free(free_tp);
	    else
	        st_free(free_tp);
	    free_tp = 0;
        }
        if(isis_ctp == &isis_scheduler && call_on_sys_stack)
        {
	    register res = (*call_on_sys_stack)(call_arg);
	    t_sig(&call_wait, res);
	    call_on_sys_stack = 0;
        }
  }

invoke(tp)
  register task *tp;
  {
        /* Initial invocation of a routine */
#ifdef	THREADS
        MUTEX_LOCK(isis_mutex);
        THREAD_SELF(tp->task_threadid);
#endif
	isis_ctp = tp;
        if(isis_state&ISIS_RECEIVING)
        {
	    isis_state &= ~ISIS_RECEIVING;
	    if(isis_gip)
	        tp->task_flag |= TASK_LOGGED;
	    if(bit(inhibit_entries, isis_enum))
	    {
	        tp->task_flag |= TASK_INHIBIT;
	        pg_join_inhibit(1);
	    }
	    if(isis_eid)
	        tp->task_eid = *isis_eid;
        }
        else
	    tp->task_entry = -1;
        if(delayed_task)
	    delayedcall(tp->task_routine, tp->task_arg0);
        else
	{
            if(isis_state&ISIS_XBYREF)
	        (*tp->task_routine)(&tp->task_arg0);
	    else
	        (*tp->task_routine)(tp->task_arg0);
	}
        /* Done, deallocate the stack */
        if((isis_state&ISIS_STARTUP) && (isis_ctp->task_flag&TASK_START))
            isis_start_done();
        qu_free(isis_ctp->task_active);
        qu_freeall(isis_ctp->task_msgs);
#ifdef  THREADS
        CONDITION_FREE(isis_ctp->task_runme);
#endif  THREADS
        if(isis_ctp->task_flag&TASK_LOGGED)
            log_end_log_msg();
        if(isis_ctp->task_flag&TASK_INHIBIT)
            pg_join_inhibit(0);
        if(isis_ctp->task_msg_arg)
            msg_delete(isis_ctp->task_msg_arg);
        free_tp = isis_ctp;
        isis_ctp = task_dequeue(&isis_runqueue);
	CONDITION_SIGNAL(isis_ctp->task_runme);
        MUTEX_UNLOCK(isis_mutex);
  }

/****************************************************************/
/*                                                              */
/*                 Basic monitor support                        */
/*                                                              */
/****************************************************************/
t_sig(cp, rval)
  register condition *cp;
  {
        register qnode *qp;
	ISIS_ENTER();
        if(qp = qu_head(*cp))
        {
            register task *np = task_dequeue(cp);
            qp->qu_task->task_rval = rval;
            task_enqueue(&isis_runqueue, np, SD);
        }
	ISIS_EXIT();
  }

t_sig_all(cp, rval)
  register condition *cp;
  {
	ISIS_ENTER();
        while(*cp)
            t_sig(cp, rval);
	ISIS_EXIT();
  }

t_sig_urgent(cp, rval)
  register condition *cp;
  {
        register qnode *qp;
	ISIS_ENTER();
        if(qp = qu_head(*cp))
        {
            qp->qu_task->task_rval = rval;
            task_enqueue(&isis_runqueue, isis_ctp, SU);
            task_swtch(task_dequeue(cp), NULLROUTINE, NULLARG);
        }
	ISIS_EXIT();
  }

#undef  t_wait
t_wait(cp)
  register condition *cp;
  {
        return(t_wait_l(cp, NULLARG));
  }


t_wait_l(cp, why)
  register condition *cp;
  char *why;
  {
        task *tp;
	ISIS_ENTER();
        if(cp != &isis_runqueue && isis_ctp == &isis_scheduler)
        {
            if(why)
                print("t_wait(%s) called from a non-task!", why);
            else
                print("t_wait(user-defined condition:%x) called from a non-task!", cp);
            ISIS_RETURN(-1);
        }
        isis_ctp->task_waitingfor = why;
        task_enqueue(cp, isis_ctp, SD);
        tp = task_dequeue(&isis_runqueue);
        task_swtch(tp, NULLROUTINE, NULLARG);
        ISIS_RETURN(isis_ctp->task_rval);
  }

t_suspend()
  {
        register task *tp;
        if(isis_ctp != &isis_scheduler)
            return(t_wait_l(&isis_runqueue, "isis system: wants to let other tasks run first"));
        tp = task_dequeue(&isis_runqueue);
        if(tp != isis_ctp)
            task_swtch(tp, NULLROUTINE, NULLARG);
        return(isis_ctp->task_rval);
  }

#ifdef THREADS

#define	MAX_ZOMBIES	10

static	nzombie;

thread_isis_enter()
  {
	register task *tp;
	register qnode *qp;
	THREAD_T self;
	if(!isis_state&ISIS_TINIT)
	    panic("ISIS task package not initialized: must call isis_init() or t_init()\n");
	THREAD_SELF(self);
	MUTEX_LOCK(isis_mutex);
	for(qp = isis_tasks->qu_next; qp != isis_tasks; qp = qp->qu_next)
	{
	    tp = qp->qu_task;
	    if(tp->task_threadid == self)
	    {
                --nzombie;
		tp->task_flag &= ~TASK_ZOMBIE;
		CONDITION_WAIT(tp->task_runme);
	        return;
	    }
	}
        tp = st_alloc();
        ++isis_created;
        tp->task_threadid = self;
        tp->task_queue = 0;
        tp->task_msgid = 0;
        tp->task_cond = 0;
        tp->task_waitingfor = 0;
        tp->task_mwant = 0;
        tp->task_msg = 0;
        tp->task_flag = TASK_TEMP;
	if(isis_state&ISIS_XBYREF)
	    tp->task_flag |= TASK_XBYREF;
        tp->task_act = isis_ctp->task_act;
        bclr(tp->task_watching);
        tp->task_active = qu_add_tp(isis_tasks, TA_ISTASK, tp);
#ifdef  CTHREADS
        tp->task_routine = (ifunc*)self->func;
        tp->task_arg0 = self->arg;
	tp->task_name = self->name;
#else   CTHREADS
        tp->task_routine = (ifunc*)0;
        tp->task_arg0 = 0;
	tp->task_name = 0;
#endif  CTHREADS
        tp->task_addr = my_address;
        tp->task_msg_arg = 0;
        tp->task_cohorts = (address*)0;
	task_enqueue(&isis_runqueue, tp, SD);
        tp->task_msgs = qu_null();
	CONDITION_ALLOC(tp->task_runme);
	CONDITION_WAIT(tp->task_runme);
  }

thread_isis_exit()
  {
	isis_ctp->task_flag |= TASK_ZOMBIE;
	if(++nzombie == MAX_ZOMBIES)
	    thread_isis_cleanup();
        isis_ctp = task_dequeue(&isis_runqueue);
        CONDITION_SIGNAL(isis_ctp->task_runme);
        MUTEX_UNLOCK(isis_mutex);
  }

thread_isis_cleanup()
  {
        register task *tp;
	register qnode *qp, *nqp;
	for(qp = isis_tasks->qu_next; qp != isis_tasks; qp = nqp)
	{
	    nqp = qp->qu_next;
	    tp = qp->qu_task;
	    if((tp->task_flag&TASK_ZOMBIE) == 0)
		continue;
            qu_free(tp->task_active);
            qu_freeall(tp->task_msgs);
	    CONDITION_FREE(tp->task_runme);
	    st_free(tp);
	}
  }
#endif THREADS
